import json
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Any

import jsonschema
import yaml
from jsonschema.validators import validate

from sbomgrader.core.enums import Grade, RuleForce, OutputType, ResultType
from sbomgrader.core.utils import get_mapping
from sbomgrader.core.definitions import (
    COOKBOOK_VALIDATION_SCHEMA_PATH,
    RULESET_DIR,
    ROOT_DIR,
    COOKBOOKS_DIR,
    COOKBOOK_EXTENSIONS,
)
from sbomgrader.core.rules import RuleSet, Document, Result, ResultDetail


@dataclass
class CookbookResult:
    result: Result
    cookbook: "Cookbook"

    def __get_by_force(self, force: RuleForce) -> Iterable[ResultDetail]:
        return [
            self.result.get(name)
            for name in getattr(self.cookbook, force.value.lower())
        ]

    @property
    def must(self) -> Iterable[ResultDetail]:
        return self.__get_by_force(RuleForce.MUST)

    @property
    def should(self) -> Iterable[ResultDetail]:
        return self.__get_by_force(RuleForce.SHOULD)

    @property
    def may(self) -> Iterable[ResultDetail]:
        return self.__get_by_force(RuleForce.MAY)

    @property
    def grade(self) -> Grade:
        unsuccessful = self.get_unsuccessful()
        if unsuccessful.must:
            return Grade.F
        grade = Grade.A
        for _ in unsuccessful.should:
            grade = Grade.lower(grade)
        return grade

    def get(self, rule_name: str):
        return self.result.get(rule_name)

    def get_unsuccessful(self) -> "CookbookResult":

        failed = self.result.failed
        error = self.result.errors
        unsuccessful = set(failed.keys())
        unsuccessful.update(error.keys())
        new_must = set(filter(lambda x: x in unsuccessful, self.cookbook.must))
        new_should = set(filter(lambda x: x in unsuccessful, self.cookbook.should))
        new_may = set(filter(lambda x: x in unsuccessful, self.cookbook.may))
        return CookbookResult(
            Result(unsuccessful, failed, error),
            Cookbook(
                self.cookbook.name,
                self.cookbook.ruleset_names,
                new_must,
                new_should,
                new_may,
            ),
        )

    def output(self, o_type: OutputType) -> str:
        if o_type in {OutputType.VISUAL, OutputType.MARKDOWN}:
            ans = f"# Cookbook: {self.cookbook.name}\n"
            ans += "\n## Summary\n"
            ans += f"\nAchieved grade: {self.grade.value}\n"
            for force in RuleForce:
                rules_in_force = self.__get_by_force(force)
                if not rules_in_force:
                    continue
                ans += f"\n### {force.value}:\n\n"
                for result_detail in rules_in_force:
                    detail = self.get(result_detail.rule_name)
                    ans += f"- {result_detail.rule_name} {ResultType.get_visual(detail.result_type)}\n"
            unsuccessful = self.get_unsuccessful()
            if unsuccessful.cookbook.all_used_rule_names:
                ans += "\n## Failure details\n\n"
                for collection in (
                    unsuccessful.result.failed,
                    unsuccessful.result.errors,
                ):
                    for rule in collection:
                        ans += f"\n### {rule}\n\n"
                        detail = self.get(rule)
                        ans += f"{detail.result_detail}\n"

            return ans
        if o_type is OutputType.JSON:
            return json.dumps(self.to_dict(), indent=4)
        return yaml.dump(self.to_dict())

    def to_dict(self) -> dict[str, Any]:
        ans = {"cookbook_name": self.cookbook.name, "grade": self.grade.value}
        for force in RuleForce:
            ans[force.value] = {}
            for result_detail in self.__get_by_force(force):
                ans[force.value][result_detail.rule_name] = {}
                ans[force.value][result_detail.rule_name][
                    result_detail.result_type.value
                ] = result_detail.result_detail
        return ans


class Cookbook:
    def __init__(
        self,
        name: str,
        ruleset_names: Iterable[str],
        must: Iterable[str],
        should: Iterable[str],
        may: Iterable[str],
    ):
        self.name = name
        self.ruleset_names = ruleset_names
        self._initialized_ruleset: RuleSet | None = None
        self.__is_initialized: bool = False
        self.must = set(must)
        self.should = set(should)
        self.may = set(may)

    @property
    def ruleset(self) -> RuleSet:
        self.initialize()
        return self._initialized_ruleset

    def __contains__(self, item):
        return item in self.must or item in self.should or item in self.may

    def __hash__(self):
        return (*self.ruleset_names, *self.must, *self.should, *self.may).__hash__()

    @property
    def all_used_rule_names(self) -> set[str]:
        selected_rules = set()
        for type_ in [self.must, self.should, self.may]:
            selected_rules.update(type_)
        return selected_rules

    def initialize(self):
        if self.__is_initialized:
            return
        self._initialized_ruleset = RuleSet()
        for ruleset in self.ruleset_names:
            if "\\" not in ruleset and "/" not in ruleset:
                # Is a native ruleset
                self._initialized_ruleset += RuleSet.from_file(
                    RULESET_DIR / (ruleset + ".yml")
                )
            else:
                # Load it from a file
                path = Path(ruleset)
                if path.is_absolute():
                    self._initialized_ruleset += RuleSet.from_file(ruleset)
                else:
                    self._initialized_ruleset += RuleSet.from_file(ROOT_DIR / ruleset)
        selected_rules = self.all_used_rule_names
        self._initialized_ruleset.selection = selected_rules
        self.__is_initialized = True

    @staticmethod
    def from_file(file_path: str | Path) -> "Cookbook":
        file_path = Path(file_path)
        schema_dict = get_mapping(file_path)
        try:
            validate(schema_dict, get_mapping(COOKBOOK_VALIDATION_SCHEMA_PATH))
        except jsonschema.exceptions.ValidationError as e:
            print(
                f"Could not parse Cookbook from file {file_path.absolute()}",
                file=sys.stderr,
            )
            raise e

        return Cookbook(
            file_path.name.rsplit(".", 1)[0],
            schema_dict["rulesets"],
            schema_dict.get(RuleForce.MUST.value, []),
            schema_dict.get(RuleForce.SHOULD.value, []),
            schema_dict.get(RuleForce.MAY.value, []),
        )

    @staticmethod
    def from_directory(dir_path: str | Path) -> list["Cookbook"]:
        dir_path = Path(dir_path)
        ans = []
        for entity in dir_path.iterdir():
            entity: Path
            if not entity.is_file():
                continue
            if not any(entity.name.endswith(ext) for ext in COOKBOOK_EXTENSIONS):
                continue
            ans.append(Cookbook.from_file(entity))
        return ans

    def __call__(self, document: dict | Document) -> CookbookResult:
        self.initialize()
        res = self._initialized_ruleset(document)
        cook_res = CookbookResult(res, self)
        return cook_res

    @staticmethod
    def load_all_defaults() -> list["Cookbook"]:
        return Cookbook.from_directory(COOKBOOKS_DIR)



import re
import sys
from dataclasses import dataclass
from typing import Union, Any, Callable

from sbomgrader.core.definitions import (
    FIELD_NOT_PRESENT,
    FieldNotPresentError,
    MAX_ITEM_PREVIEW_LENGTH,
    START_PREVIEW_CHARS,
    END_PREVIEW_CHARS,
)
from sbomgrader.core.enums import QueryType


class PathParser:
    def __init__(self, path: str):
        self._path = path
        self.char_no = 0
        self.next_is_query = False

    def __create_field(
        self, field: str, next_is_query: bool
    ) -> Union[str, "QueryParser"]:
        if self.next_is_query:
            next_ = QueryParser(field)
        else:
            next_ = field.strip()
        self.next_is_query = next_is_query
        return next_

    def next(self) -> Union[str, list["QueryParser"], None]:
        in_block = 1 if self.next_is_query else 0
        buffer = ""
        for char in self._path[self.char_no :]:
            self.char_no += 1
            if char == "[":
                if not in_block:
                    return self.__create_field(buffer, True)
                buffer += char
                in_block += 1
            elif char == "]":
                in_block -= 1
                if not in_block:
                    return self.__create_field(buffer, False)
                buffer += char
            elif char == ".":
                if not in_block:
                    return self.__create_field(buffer, False)
                buffer += char
            else:
                buffer += char
        if buffer:
            return self.__create_field(buffer, False)
        return None

    @property
    def all(self) -> list[Union[str, "QueryParser"]]:
        backup = self.char_no
        self.char_no = 0
        ans = []
        step = self.next()
        while step is not None:
            ans.append(step)
            step = self.next()
        self.char_no = backup
        return ans


@dataclass
class Query:
    type_: QueryType
    value: str | None
    field_path: PathParser | None

    @property
    def variable(self) -> str | None:
        if self.value and (match := re.match(r"^\$\{(?P<varname>\w+)\}$", self.value)):
            return match.group("varname")


class QueryParser:
    def __init__(self, path: str):
        self._path = path

    def parse(self) -> list[Query]:
        queries = []
        field_buffer = ""
        operation_buffer = ""
        value_buffer = ""
        in_block = 0
        after_operation = False
        for char in self._path:
            if re.match(r"\s", char) and not after_operation:
                continue
            if char in {"!", "=", "%", "|", "&"} and not in_block:
                operation_buffer += char
                after_operation = True
            elif after_operation and char != ",":
                value_buffer += char
            elif char == "," and after_operation:
                queries.append(
                    Query(
                        type_=QueryType(operation_buffer),
                        field_path=(
                            None if not field_buffer else PathParser(field_buffer)
                        ),
                        value=None if not value_buffer else value_buffer,
                    )
                )
                field_buffer = ""
                operation_buffer = ""
                value_buffer = ""
                after_operation = False
            elif char == "[":
                field_buffer += char
                in_block += 1
            elif char == "]":
                in_block -= 1
                field_buffer += char

            else:
                field_buffer += char.strip()

        if field_buffer or operation_buffer or value_buffer:
            queries.append(
                Query(
                    type_=QueryType(operation_buffer.strip()),
                    field_path=PathParser(field_buffer.strip()),
                    value=value_buffer.strip(),
                )
            )
        return queries


class FieldResolver:

    def __init__(self, variables: dict[str, str]):
        self._uninitialized_vars = variables

    def has_var(self, var_name: str) -> bool:
        return var_name in self._uninitialized_vars

    @property
    def var_definitions(self) -> dict[str, str]:
        return self._uninitialized_vars

    def resolve_variables(self, doc: dict[str, Any]) -> dict[str, set]:
        # first resolve dependency tree for variables
        dependencies = {}
        for variable_name, variable_path in self._uninitialized_vars.items():
            dependencies[variable_name] = set()
            dependencies[variable_name].update(
                match.group("varname")
                for match in re.finditer(r"\${(?P<varname>\w+)}", variable_path)
            )
            assert (
                variable_name not in dependencies[variable_name]
            ), f"Self referencing variable {variable_name} found."
        resolved_variables: dict[str, set] = {}
        while not all(var_name in resolved_variables for var_name in dependencies):
            # Get a var with no dependencies
            var_name, var_deps = sorted(dependencies.items(), key=lambda x: len(x[1]))[
                0
            ]
            assert (
                not var_deps
            ), f"Circular variable reference found for variable {var_name}"

            resolved_variables[var_name] = set()

            def add_to_variable(value: Any) -> None:
                resolved_variables[var_name].add(value)

            path = PathParser(self._uninitialized_vars[var_name]).all
            try:
                self._run_on_path(
                    doc,
                    path,
                    resolved_variables,
                    "",
                    add_to_variable,
                    False,
                    set(),
                )
            except Exception as e:
                print(
                    f"Could not parse variable {var_name}, problem: {str(e)}",
                    file=sys.stderr,
                )

            dependencies.pop(var_name)
            for dep in dependencies.values():
                if var_name in dep:
                    dep.remove(var_name)
        return resolved_variables

    def _run_on_path(
        self,
        doc_: Union[dict, list[Any], FIELD_NOT_PRESENT],
        path: list[str | QueryParser | PathParser],
        variables: dict[str, Any],
        path_tried: str,
        func_to_run: Callable[[Any], Any],
        accept_not_present_field: bool,
        ran_on: set[str],
    ):

        if not accept_not_present_field and doc_ is FIELD_NOT_PRESENT:
            raise FieldNotPresentError("Field not present: ", path_tried)
        if not path:
            # The path has ended
            try:
                resp = func_to_run(doc_)
                ran_on.add(path_tried)
                assert resp is True or resp is None
            except Exception as e:
                item_str = str(doc_)
                if len(item_str) > MAX_ITEM_PREVIEW_LENGTH:
                    item_str = f"{item_str[:START_PREVIEW_CHARS]}...{item_str[-END_PREVIEW_CHARS:]}"
                if not path_tried:
                    path_tried = "."
                message_to_return = (
                    f"Check did not pass for item: {item_str} at path: {path_tried}\n"
                    + "\n".join(str(m) for m in e.args)
                )
                raise type(e)(message_to_return)
            return
        step = path[0]
        if isinstance(step, str):
            # Field name
            assert isinstance(
                doc_, dict
            ), f"Cannot access field '{step}' on other objects than dicts. Provided object: {doc_}"
            if step == "?":
                assert path[1:] and isinstance(
                    path[1], str
                ), "Cannot use ? before anything else than a field name."
                if path[1] in doc_:
                    self._run_on_path(
                        doc_,
                        path[1:],
                        variables,
                        path_tried,
                        func_to_run,
                        accept_not_present_field,
                        ran_on,
                    )
            else:
                self._run_on_path(
                    doc_.get(step, FIELD_NOT_PRESENT),
                    path[1:],
                    variables,
                    path_tried + f".{step}",
                    func_to_run,
                    accept_not_present_field,
                    ran_on,
                )
        elif isinstance(step, QueryParser):
            assert isinstance(
                doc_, list
            ), f"Queries can only be performed on lists! Tested path: {path_tried}, item: {doc_}"
            queries = step.parse()

            to_use = []
            can_fail_for_some = False

            for query in queries:
                if query.type_ in {QueryType.EACH, QueryType.ANY}:
                    # Use every list index available
                    to_use.append(set(range(len(doc_))))
                    if query.type_ is QueryType.ANY:
                        can_fail_for_some = True
                    continue
                # Actually filter the list
                to_use_in_query = set()
                for idx, item in enumerate(doc_):
                    varname = query.variable
                    if query.type_ is QueryType.EQ:
                        if varname:
                            func = lambda x: x in variables[varname]
                        else:
                            func = lambda x: str(x) == query.value
                    elif query.type_ is QueryType.NEQ:
                        if varname:
                            func = lambda x: x not in variables[varname]
                        else:
                            func = lambda x: str(x) != query.value
                    elif query.type_ is QueryType.STARTSWITH:
                        if varname:
                            func = lambda x: isinstance(x, str) and any(
                                x.startswith(val) for val in variables[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and x.startswith(
                                query.value
                            )
                    elif query.type_ is QueryType.ENDSWITH:
                        if varname:
                            func = lambda x: isinstance(x, str) and any(
                                x.endswith(val) for val in variables[varname]
                            )
                        else:
                            func = lambda x: isinstance(x, str) and x.endswith(
                                query.value
                            )
                    final_func = lambda x: (
                        to_use_in_query.add(idx) if func(x) else None
                    )
                    self._run_on_path(
                        item,
                        query.field_path.all,
                        variables,
                        path_tried + f"[{idx}]",
                        final_func,
                        True,
                        set(),
                    )
                    to_use.append(to_use_in_query)
            to_use_final = set.intersection(*to_use) if to_use else {}
            failed = 0
            assertions = []
            for idx, item in enumerate(doc_):
                if idx not in to_use_final:
                    continue

                if can_fail_for_some:
                    try:
                        self._run_on_path(
                            item,
                            path[1:],
                            variables,
                            path_tried + f"[{idx}]",
                            func_to_run,
                            accept_not_present_field,
                            ran_on,
                        )
                    except (AssertionError, FieldNotPresentError) as e:
                        failed += 1
                        assertions.append(e)
                    assert failed < len(
                        to_use_final
                    ), f"Check did not pass for any fields. Assertions: {assertions}, path: {path_tried}"
                else:
                    self._run_on_path(
                        item,
                        path[1:],
                        variables,
                        path_tried + f"[{idx}]",
                        func_to_run,
                        accept_not_present_field,
                        ran_on,
                    )

    def run_func(
        self,
        doc: dict[str, Any],
        func: Callable[[Any], Any],
        field_path: str,
        minimal_runs: int = 1,
        fallback_variables: dict[str, Any] | None = None,
    ) -> Any:
        variables = {} if not fallback_variables else {**fallback_variables}
        variables.update(self.resolve_variables(doc))

        ran_on = set()
        self._run_on_path(
            doc, PathParser(field_path).all, variables, "", func, False, ran_on
        )
        assert (
            len(ran_on) >= minimal_runs
        ), "Test was not performed on any fields because no fields match given filters."



import pytest

from sbomgrader.core.cookbooks import Cookbook
from sbomgrader.core.documents import Document


@pytest.mark.parametrize(
    ["sbom_fixture_name", "cookbook_fixture_name"],
    [
        ("image_build_sbom", "image_build_cookbook"),
        ("image_release_sbom", "image_release_cookbook"),
        ("image_index_build_sbom", "image_index_build_cookbook"),
        ("image_index_release_sbom", "image_index_release_cookbook"),
        ("product_sbom", "product_cookbook"),
        ("rpm_build_sbom", "rpm_build_cookbook"),
        ("rpm_release_sbom", "rpm_release_cookbook"),
    ],
)
def test_success_cookbook_build(sbom_fixture_name, cookbook_fixture_name, request):
    cookbook: Cookbook = request.getfixturevalue(cookbook_fixture_name)
    sbom_doc: Document = request.getfixturevalue(sbom_fixture_name)
    res = cookbook(sbom_doc)
    assert res.result.ran
    assert not res.result.not_implemented
    unsuccessful = res.get_unsuccessful()
    assert not unsuccessful.may
    assert not unsuccessful.should
    assert not unsuccessful.must
    assert not res.result.not_implemented



import re

from packageurl import PackageURL


def _get_relationships(doc: dict) -> list:
    relationships = doc.get("relationships")
    assert relationships, "Missing field 'relationships'"
    return relationships


def _get_main_packages(doc: dict) -> list:
    relationships = _get_relationships(doc)
    main_element_relationship = list(
        filter(
            lambda x: x.get("spdxElementId") == "SPDXRef-DOCUMENT"
            and x.get("relationshipType") == "DESCRIBES",
            relationships,
        )
    )
    main_packages = []
    for main_element in main_element_relationship:
        expected_spdxid = main_element.get("relatedSpdxElement")
        referenced_package = next(
            filter(
                lambda x: x.get("SPDXID") == expected_spdxid, doc.get("packages", [])
            )
        )
        main_packages.append(referenced_package)
    return main_packages


def image_packages_variants(doc: dict):
    main_package_SPDXIDs = {p.get("SPDXID") for p in _get_main_packages(doc)}
    for package in doc.get("packages", []):
        if package["SPDXID"] in main_package_SPDXIDs:
            continue
        assert next(
            filter(
                lambda x: x.get("relationshipType") == "VARIANT_OF"
                and x.get("spdxElementId") == package["SPDXID"]
                and x.get("relatedSpdxElement") in main_package_SPDXIDs,
                doc.get("relationships", []),
            ),
            None,
        ), f"Package {package["SPDXID"]} is not variant of main element."


def purl_has_repo_and_tag_qualifiers(purl: str):
    purl = PackageURL.from_string(purl)
    assert "repository_url" in purl.qualifiers
    assert "tag" in purl.qualifiers



import json
from pathlib import Path

import yaml

from sbomgrader.core.enums import Grade


def get_mapping(schema: str | Path) -> dict | None:
    if isinstance(schema, str):
        schema = Path(schema)
    if isinstance(schema, Path):
        if not schema.exists():
            return None
        with open(schema) as stream:
            if schema.name.endswith(".json"):
                doc = json.load(stream)
            elif schema.name.endswith(".yml") or schema.name.endswith(".yaml"):
                doc = yaml.safe_load(stream)
            else:
                doc = {}
            return doc


def get_path_to_implementations(schema_path: str | Path):
    if isinstance(schema_path, str):
        schema_path = Path(schema_path)
    return schema_path.parent / "implementations" / schema_path.name.rsplit(".", 1)[0]


def validation_passed(validation_grade: Grade, minimal_grade: Grade) -> bool:
    # minimal is less than or equal to validation
    return Grade.compare(validation_grade, minimal_grade) < 1



import re
from pathlib import Path
from typing import Sized

from sbomgrader.core.enums import Implementation

ROOT_DIR: Path = Path(__file__).parent.parent
RULESET_DIR = ROOT_DIR / "rulesets"
COOKBOOKS_DIR = ROOT_DIR / "cookbooks"
IMPLEMENTATION_DIR_NAME = "specification_rules"
RULESET_VALIDATION_SCHEMA_PATH = ROOT_DIR / "rulesets" / "schema" / "rule_schema.yml"
COOKBOOK_VALIDATION_SCHEMA_PATH = (
    ROOT_DIR / "cookbooks" / "schema" / "cookbook_schema.yml"
)
COOKBOOK_EXTENSIONS = {".yml", ".yaml"}

SBOM_FORMAT_DEFINITION_MAPPING = {
    Implementation.SPDX23: {"spdxVersion": "SPDX-2.3"},
    Implementation.CYCLONEDX15: {"bomFormat": "CycloneDX", "specVersion": "1.5"},
}
MAX_ITEM_PREVIEW_LENGTH = 50
START_PREVIEW_CHARS = 25
END_PREVIEW_CHARS = 20


class __FieldNotPresent:
    def __repr__(self):
        return "Field not present."

    def get(self, *_):
        return self


FIELD_NOT_PRESENT = __FieldNotPresent()


class FieldNotPresentError(ValueError):
    pass


operation_map = {
    "eq": lambda expected, actual: expected == actual,
    "neq": lambda expected, actual: expected != actual,
    "in": lambda expected, actual: actual in expected,
    "not_in": lambda expected, actual: actual not in expected,
    "str_startswith": lambda expected, actual: isinstance(actual, str)
    and actual.startswith(expected),
    "str_endswith": lambda expected, actual: isinstance(actual, str)
    and actual.endswith(expected),
    "str_contains": lambda expected, actual: isinstance(actual, str)
    and expected in actual,
    "str_matches_regex": lambda expected, actual: isinstance(actual, str)
    and bool(re.match(expected, actual)),
    "length_eq": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) == expected,
    "length_gt": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) > expected,
    "length_lt": lambda expected, actual: isinstance(actual, Sized)
    and len(actual) < expected,
    "func_name": None,
}



from collections import defaultdict
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Any, Callable

from jsonschema.validators import validate

from sbomgrader.core.documents import Document
from sbomgrader.core.enums import ResultType
from sbomgrader.core.field_resolve import FieldResolver
from sbomgrader.core.rule_loader import RuleLoader
from sbomgrader.core.definitions import (
    RULESET_VALIDATION_SCHEMA_PATH,
    FIELD_NOT_PRESENT,
    FieldNotPresentError,
    operation_map,
)
from sbomgrader.core.utils import get_mapping, get_path_to_implementations


@dataclass
class ResultDetail:
    rule_name: str
    result_type: ResultType
    result_detail: str | None


@dataclass
class Result:
    ran: set[str] = field(default_factory=set)
    failed: dict[str, str] = field(default_factory=dict)
    errors: dict[str, str] = field(default_factory=dict)
    skipped: set[str] = field(default_factory=set)
    not_implemented: set[str] = field(default_factory=set)

    def __add__(self, other: "Result") -> "Result":
        if not isinstance(other, Result):
            raise TypeError(f"Cannot add Result and {type(other)}")
        return Result(
            failed=self.failed | other.failed,
            ran=self.ran | other.ran,
            errors=self.errors | other.errors,
            skipped=self.skipped | other.skipped,
            not_implemented=self.not_implemented | other.not_implemented,
        )

    def get(self, rule_name: str) -> ResultDetail:
        if rule_name in self.failed:
            return ResultDetail(
                rule_name=rule_name,
                result_type=ResultType.FAILED,
                result_detail=self.failed[rule_name],
            )
        if rule_name in self.errors:
            return ResultDetail(
                rule_name=rule_name,
                result_type=ResultType.ERROR,
                result_detail=self.errors[rule_name],
            )
        if rule_name in self.skipped:
            return ResultDetail(
                rule_name=rule_name,
                result_type=ResultType.SKIPPED,
                result_detail="Rule was not present in the cookbook.",
            )
        if rule_name in self.not_implemented:
            return ResultDetail(
                rule_name=rule_name,
                result_type=ResultType.NOT_IMPLEMENTED,
                result_detail="No implementation found for the document type.",
            )
        if rule_name in self.ran:
            return ResultDetail(
                rule_name=rule_name,
                result_type=ResultType.SUCCESS,
                result_detail="Success.",
            )
        return ResultDetail(
            rule_name=rule_name,
            result_type=ResultType.NOT_PRESENT,
            result_detail="Rule is not present in any RuleSet.",
        )


@dataclass
class Rule:
    name: str
    func: Callable
    error_message: str
    field_path: str
    minimum_tested_elements: int
    field_resolver: FieldResolver

    def __call__(
        self,
        doc: list[dict] | dict | Document,
        fallback_vars: dict[str, Any] | None = None,
    ) -> Result:
        if isinstance(doc, dict):
            doc = Document(doc)

        result = Result(ran={self.name})
        field_path = self.field_path or ""
        fallback_vars = {} if not fallback_vars else fallback_vars
        try:
            self.field_resolver.run_func(
                doc.doc,
                self.func,
                field_path,
                self.minimum_tested_elements,
                fallback_variables=fallback_vars,
            )

        except AssertionError as e:
            message_to_return = self.error_message
            if e.args:
                message_to_return += "\nDetail from runtime: " + "\n".join(
                    str(m) for m in e.args
                )
            result.failed[self.name] = message_to_return
        except FieldNotPresentError as e:
            result.failed[self.name] = (
                self.error_message + " Field not present: " + e.args[1]
            )
        except Exception as e:
            result.errors[self.name] = str(type(e)) + " " + str(e)
        return result


class RuleSet:
    @staticmethod
    def from_file(file: str | Path):
        schema_dict = get_mapping(file)

        validate(schema_dict, get_mapping(RULESET_VALIDATION_SCHEMA_PATH))

        implementation_loaders: dict[str, RuleLoader] = {}

        for implementation_file in get_path_to_implementations(file).iterdir():
            implementation_name = implementation_file.name.rsplit(".", 1)[0]
            implementation_loaders[implementation_name] = RuleLoader(
                implementation_name, implementation_file
            )

        global_variable_definitions = {}
        for implementation_obj in schema_dict.get("variables", {}).get(
            "implementations", []
        ):

            implementation = implementation_obj["name"]
            global_variable_definitions[implementation] = {}
            for var_obj in implementation_obj["variables"]:
                global_variable_definitions[implementation][var_obj["name"]] = var_obj[
                    "fieldPath"
                ]

        all_rules = defaultdict(dict)
        for rule in schema_dict["rules"]:
            name = rule["name"]
            implementations = rule["implementations"]
            failure_message = rule["failureMessage"]
            for spec in implementations:
                implementation_name = spec["name"]
                field_path = spec.get("fieldPath")
                checker = spec["checker"]
                # TODO rework for more than one operation
                operation = next(iter(checker.keys()))

                check_against = checker[operation]
                if check_against == "FIELD_NOT_PRESENT":
                    check_against = FIELD_NOT_PRESENT
                elif isinstance(check_against, list):
                    contains_field_not_present = "FIELD_NOT_PRESENT" in check_against
                    check_against = [
                        item for item in check_against if item != "FIELD_NOT_PRESENT"
                    ]
                    if contains_field_not_present:
                        check_against.append(FIELD_NOT_PRESENT)
                func = operation_map[operation]

                if not callable(func):
                    # load func according to name
                    func = implementation_loaders[implementation_name].load_rule(
                        check_against
                    )
                else:
                    func = partial(func, check_against)

                var_dict = {}
                spec_variables = spec.get("variables", [])
                for var_obj in spec_variables:
                    var_dict[var_obj["name"]] = var_obj["fieldPath"]

                minimum_tested_elements = spec.get("minimumTestedElements", 1)
                failure_message = spec.get("failureMessage") or failure_message

                all_rules[implementation_name][name] = Rule(
                    name=name,
                    func=func,
                    error_message=failure_message,
                    field_path=field_path,
                    field_resolver=FieldResolver(var_dict),
                    minimum_tested_elements=minimum_tested_elements,
                )
        all_rule_names = {rule["name"] for rule in schema_dict["rules"]}
        return RuleSet(
            implementation_loaders=implementation_loaders,
            rules=all_rules,
            all_rule_names=all_rule_names,
            variables=global_variable_definitions,
        )

    def __init__(
        self,
        rules: dict[str, dict[str, Callable]] | None = None,
        implementation_loaders: dict[str, RuleLoader] | None = None,
        all_rule_names: set[str] | None = None,
        selection: set[str] | None = None,
        variables: dict[str, dict[str, str]] | None = None,
    ):
        self.implementation_loaders = implementation_loaders or {}
        self.rules = rules or {}
        self.all_rule_names = all_rule_names or set()
        self.selection = selection if selection is not None else self.all_rule_names
        self.field_resolvers = {}
        if not variables:
            variables = {}
        for implementation, var_dict in variables.items():
            self.field_resolvers[implementation] = FieldResolver(var_dict)

    def __add__(self, other: "RuleSet"):
        if not isinstance(other, RuleSet):
            raise TypeError(f"Cannot combine RuleSet with instance of {type(other)}!")
        implementation_loaders = {}
        for implementation in [
            *self.implementation_loaders.keys(),
            *other.implementation_loaders.keys(),
        ]:
            new_loader = RuleLoader(implementation)
            self_loader = self.implementation_loaders.get(implementation, None)
            other_loader = other.implementation_loaders.get(implementation, None)
            if self_loader:
                new_loader.add_file_references(*self_loader.file_references)
            if other_loader:
                new_loader.add_file_references(*other_loader.file_references)
            implementation_loaders[implementation] = new_loader

        new_rules = {}
        for implementation in [*self.rules.keys(), *other.rules.keys()]:
            new_rules[implementation] = {
                **self.rules.get(implementation, {}),
                **other.rules.get(implementation, {}),
            }
        variable_definitions = {}
        for implementation in [*self.field_resolvers.keys(), *other.rules.keys()]:
            variable_definitions[implementation] = {}
            self_resolver = self.field_resolvers.get(implementation)
            if self_resolver:
                variable_definitions[implementation].update(
                    self_resolver.var_definitions
                )
            other_resolver = other.field_resolvers.get(implementation)
            if other_resolver:
                variable_definitions[implementation].update(
                    other_resolver.var_definitions
                )
        return RuleSet(
            rules=new_rules,
            all_rule_names=self.all_rule_names | other.all_rule_names,
            selection=self.selection | other.selection,
            implementation_loaders=implementation_loaders,
            variables=variable_definitions,
        )

    def __call__(self, document: dict | Document) -> Result:
        res = Result()
        if isinstance(document, dict):
            document = Document(document)
        implementation = document.implementation.value
        global_variables_resolver = self.field_resolvers.get(implementation)
        global_variables = {}
        if global_variables_resolver:
            global_variables = global_variables_resolver.resolve_variables(document.doc)

        for rule in self.all_rule_names:
            if rule not in self.selection:
                res.skipped.add(rule)
                continue
            if rule_obj := self.rules.get(implementation, {}).get(rule):
                rule_obj: Rule | None
                if not callable(rule_obj.func):
                    res.not_implemented.add(rule)
                else:
                    res += rule_obj(document, fallback_vars=global_variables)
            else:
                res.not_implemented.add(rule)
        return res



from copy import copy
from functools import cached_property
from typing import Any, Iterable

from sbomgrader.core.definitions import SBOM_FORMAT_DEFINITION_MAPPING
from sbomgrader.core.enums import SBOMType, Implementation


class Document:
    def __init__(self, document_dict: dict[str, Any]):
        self._doc = document_dict

    @cached_property
    def implementation(self) -> Implementation:
        for item in Implementation:
            field_to_check = SBOM_FORMAT_DEFINITION_MAPPING[item]

            if all(
                self._doc.get(key) == value for key, value in field_to_check.items()
            ):
                return item
        raise NotImplementedError("Document is in an unsupported standard.")

    @property
    def sbom_type(self) -> "SBOMType":
        if self.implementation is Implementation.SPDX23:
            relationships = self._doc.get("relationships", [])
            main_relationships = [
                relationship
                for relationship in relationships
                if relationship["spdxElementId"] == "SPDXRef-DOCUMENT"
                and relationship["relationshipType"] == "DESCRIBES"
            ]
            if len(main_relationships) > 1:
                raise ValueError(
                    "Cannot determine single SBOMType from multi-sbom. Try separating docs first."
                )
            main_relationship = main_relationships[0]
            main_spdxid = main_relationship["relatedSpdxElement"]
            first_degree_relationships = [
                relationship
                for relationship in relationships
                if (
                    relationship["spdxElementId"] == main_spdxid
                    or relationship["relatedSpdxElement"] == main_spdxid
                )
                and relationship != main_relationship
            ]
            if all(
                relationship["relationshipType"] == "VARIANT_OF"
                for relationship in first_degree_relationships
            ):
                return SBOMType.IMAGE_INDEX
            if all(
                relationship["relationshipType"]
                in {"DESCENDANT_OF", "CONTAINS", "BUILD_TOOL_OF"}
                for relationship in first_degree_relationships
            ):
                return SBOMType.IMAGE
            if all(
                relationship["relationshipType"] in {"GENERATED_FROM", "CONTAINS"}
                for relationship in first_degree_relationships
            ):
                return SBOMType.RPM

            def sort_relationship_key(relationship: dict):
                return "".join(sorted(relationship.values()))

            if sorted(
                first_degree_relationships + main_relationships,
                key=sort_relationship_key,
            ) == sorted(relationships, key=sort_relationship_key):
                return SBOMType.PRODUCT
            return SBOMType.UNKNOWN
        elif self.implementation is Implementation.CYCLONEDX15:
            if self._doc.get("metadata", {}).get("component", {}).get("type") in {
                "operating-system"
            }:
                return SBOMType.PRODUCT
            return SBOMType.UNKNOWN
        else:
            raise NotImplementedError()

    @property
    def doc(self):
        return self._doc



import sys
from argparse import ArgumentParser
from pathlib import Path

from rich.console import Console
from rich.markdown import Markdown

from sbomgrader.core.cookbook_bundles import CookbookBundle
from sbomgrader.core.cookbooks import Cookbook
from sbomgrader.core.documents import Document
from sbomgrader.core.enums import Grade, SBOMTime, OutputType, SBOMType
from sbomgrader.core.utils import get_mapping, validation_passed


def main():
    parser = ArgumentParser("sbomgrader")
    parser.add_argument(
        "input",
        type=Path,
        help="SBOM File to grade. Currently supports JSON.",
        nargs="?",
    )
    parser.add_argument(
        "--cookbooks",
        "-c",
        action="append",
        type=str,
        help="Cookbooks to use for validation. "
        "Might reference default cookbooks, directories or files. "
        "Only files with '.yml' or '.yaml' extensions are taken into account if files or directories are specified.",
    )
    parser.add_argument(
        "--list-cookbooks",
        "-l",
        action="store_true",
        default=False,
        help="List available default cookbooks and exit.",
    )
    parser.add_argument(
        "--content-type",
        "-ct",
        choices=[v.value for v in SBOMType if v is not SBOMType.UNSPECIFIED],
        default=SBOMType.UNSPECIFIED.value,
        help="Specify SBOM content type. Ignored if cookbooks argument is specified.",
    )
    parser.add_argument(
        "--sbom-type",
        "-st",
        choices=[v.value for v in SBOMTime if v is not SBOMTime.UNSPECIFIED],
        default=None,
        help="If using the standard validation, specify which SBOM type (by time) is being validated. "
        "Ignored if cookbooks argument is specified.",
    )
    parser.add_argument(
        "--passing-grade",
        "-g",
        choices=[v.value for v in Grade],
        default=Grade.B.value,
        help="Minimal passing grade. Default is B.",
    )
    parser.add_argument(
        "--output",
        "-o",
        choices=[v.value for v in OutputType],
        default=OutputType.VISUAL.value,
        help="Specify the output format.",
    )

    args = parser.parse_args()
    console = Console()
    default_cookbooks = Cookbook.load_all_defaults()
    if args.list_cookbooks:
        console.print(Markdown("\n".join(f"- {cb.name}" for cb in default_cookbooks)))
        exit(0)

    sbom_file = args.input
    if not sbom_file:
        print("Please supply an SBOM file.", file=sys.stderr)
        parser.print_help(sys.stderr)
        exit(1)
    doc = Document(get_mapping(sbom_file))

    cookbook_bundles = []
    if args.cookbooks:
        cookbook_bundle = CookbookBundle([])
        for cookbook in args.cookbooks:
            cookbook_obj = next(
                filter(lambda x: x.name == cookbook, default_cookbooks), None
            )
            if cookbook_obj:
                # It's a default cookbook name
                cookbook_bundle += cookbook_obj
                continue
            cookbook = Path(cookbook)
            if cookbook.is_dir():
                cookbook_bundle += CookbookBundle.from_directory(cookbook)
                if not cookbook_bundle.cookbooks:
                    print(
                        f"Could not find any cookbooks in directory {cookbook.absolute()}",
                        file=sys.stderr,
                    )
            elif cookbook.is_file() and (
                cookbook.name.endswith(".yml") or cookbook.name.endswith(".yaml")
            ):
                cookbook_bundles.append(CookbookBundle([Cookbook.from_file(cookbook)]))
            else:
                print(f"Could not find cookbook {cookbook.absolute()}", file=sys.stderr)

        for cb in cookbook_bundles:
            cookbook_bundle += cb
        if not cookbook_bundle.cookbooks:
            print("No cookbook(s) could be found.", file=sys.stderr)
            exit(1)
    else:
        # Cookbooks weren't specified, using defaults
        type_ = SBOMType(args.type)
        if type_ is SBOMType.UNSPECIFIED:
            type_ = doc.sbom_type
        cookbook_bundle = CookbookBundle.for_document_type(type_, SBOMTime(args.time))

    result = cookbook_bundle(doc)

    output_type = OutputType(args.output)
    if output_type is OutputType.VISUAL:
        markdown = Markdown(result.output(output_type))
        console.print(markdown)
    else:
        console.print(result.output(output_type), output_type.value)
    if validation_passed(result.grade, Grade(args.passing_grade)):
        exit(0)
    exit(1)


if __name__ == "__main__":
    main()
